import time
import asyncio
from pymetawear.discover import discover_devices
from pymetawear.client import MetaWearClient
import websockets
import json
import functools
from threading import Thread
from multiprocessing import Process, Queue, cpu_count
import signal
import os
import pandas as pd
import time
import random
import concurrent.futures
from datetime import datetime
from .input_stream import InputStream
import json
import copy
import math
from .stream_package import SensorStreamPackage, SensorStreamChunk
from collections import deque
from functools import partial
import logging


class ChunkMerger(object):
    def __init__(self, data_type, n_streams, update_rate):
        self._data_type = data_type
        self._chunk_buffer = {}
        self._next_chunk_st = None
        self._merged = []
        self._n_streams = n_streams
        self._update_rate = update_rate
        self._merged_chunk_index = 0

    def process_chunk(self, chunk, send_callback):
        result = None
        self._add_chunk(chunk)
        self._merged += self._look_in_buffer(self._next_chunk_st)
        if self._is_time_to_merge(chunk):
            result = self._merge()
            send_callback(result)
            # next merge
            self._next_chunk_st = self._next_chunk_st + self._update_rate

    def _look_in_buffer(self, st):
        matched = []
        for id, buffer in self._chunk_buffer.items():
            while len(buffer) > 0 and buffer[0].get_chunk_st() < st:
                print('Discard: ' + buffer[0].get_stream_name() +
                      ' at start time: ' + str(buffer[0].get_chunk_st()))
                buffer.popleft()
            if len(buffer) > 0 and buffer[0].get_chunk_st == st:
                matched.append(buffer.popleft())
        return matched

    def _add_chunk(self, chunk):
        if self._next_chunk_st is None:
            # create a new merge
            print('add to merge')
            print(chunk)
            self._next_chunk_st = chunk.get_chunk_st()
            self._merged.append(chunk)
        else:
            if chunk.get_chunk_st() < self._next_chunk_st:
                print('discard this chunk from past from merger')
            elif chunk.get_chunk_st() == self._next_chunk_st:
                print('add to merge')
                print(chunk)
                self._merged.append(chunk)
            else:
                print('add to buffer')
                print(chunk)
                self._add_to_buffer(chunk)

    def _add_to_buffer(self, chunk):
        id = chunk.get_device_id()
        if id in self._chunk_buffer:
            self._chunk_buffer[id].append(chunk)
        else:
            self._chunk_buffer[id] = deque()
            self._chunk_buffer[id].append(chunk)

    def _is_time_to_merge(self, chunk):
        id = chunk.get_device_id()
        # if we have data from every stream
        if len(self._merged) == self._n_streams:
            return True
        # if we have waited for more than two chunks from the same stream
        elif id in self._chunk_buffer and len(
                self._chunk_buffer[id]) > 2 and len(self._merged) > 0:
            print('size of ' + chunk.get_stream_name() + ': ' +
                  str(len(self._chunk_buffer[id])))
            return True
        return False

    def _merge(self):
        self._merged.sort(key=lambda chunk: chunk.get_stream_order())
        stream_names = [stream.get_stream_name() for stream in self._merged]
        result = {
            'DATA_TYPE': self._merged[0].get_data_type(),
            'N_STREAMS': len(self._merged),
            'STREAM_NAMES': stream_names,
            'CHUNKS': copy.deepcopy(self._merged),
            'MERGED_CHUNK_ST': self._merged[0].get_chunk_st(),
            'MERGED_CHUNK_ET': self._merged[0].get_chunk_et(),
            'MERGED_CHUNK_INDEX': self._merged_chunk_index
        }
        print(result['DATA_TYPE'] + ' merged chunk ' +
              str(result['MERGED_CHUNK_INDEX']) + ' includes ' +
              ','.join(result['STREAM_NAMES']) + ' from ' +
              str(result['MERGED_CHUNK_ST']) + ' to ' +
              str(result['MERGED_CHUNK_ET']))
        self._merged_chunk_index += 1
        self._merged.clear()
        return result


class ProcessorStream(object):
    def __init__(self,
                 loop,
                 allow_ws=True,
                 host='*',
                 port=9000,
                 number_of_windows=None):
        self._host = host
        self._port = port
        self._data_queue = asyncio.Queue(loop=loop)
        self._result_queue = asyncio.Queue(loop=loop)
        self._clients = set()
        self._start_ws_server = websockets.serve(
            self._ws_handler, host=host, port=port, loop=loop)
        self._loop = loop
        self._future_results = set()
        self._allow_ws = allow_ws
        self._stopped = False
        self._ws_server_stopped = False
        self._number_of_windows = number_of_windows
        self._result_counter = 0

    def get_ws_url(self):
        return 'ws://' + self._host + ':' + str(self._port)

    def put_input_queue(self, data):
        self._loop.call_soon_threadsafe(self._data_queue.put_nowait, data)

    def put_output_queue(self, data):
        def _put_into_queue():
            self._result_queue.put_nowait(data)
            self._result_counter += 1

        self._loop.call_soon_threadsafe(_put_into_queue)

    def get_input_queue_size(self):
        return self._data_queue.qsize()

    def get_output_queue_size(self):
        return self._result_queue.qsize()

    def set_processor_pipeline(self, pipeline, **kwargs):
        func = partial(pipeline, **kwargs)
        self._pipeline = {'func': func, 'last_run': 0}

    def _sending_result(self, future):
        result = future.result()
        self._future_results.remove(future)
        # print('Remaining jobs: ' + str(len(self._future_results)))
        # print('computed data: ' + str(result[0]['HEADER_TIME_STAMP']))
        # if len(self._clients) > 0:
        if self._allow_ws and result is not None:
            self.put_output_queue(result)
            self._loop.call_soon_threadsafe(
                print, 'result_queue size: ' + str(self._result_queue.qsize()))

    async def _do_computing(self):
        sleep_time = 2
        with concurrent.futures.ProcessPoolExecutor(
                max_workers=int(cpu_count() / 2)) as pool:
            while True:
                data = await self._data_queue.get()
                if data == 'STOP':
                    if len(self._clients) > 0:
                        self.put_output_queue(data)
                        print('Stop in do computing')
                    else:
                        print('Stop in do computing')
                        self._stopped = True
                    break
                current_ts = time.time()
                print(current_ts - self._pipeline['last_run'])
                if self._pipeline['last_run'] == 0:
                    self._pipeline['last_run'] = current_ts
                elif current_ts - self._pipeline['last_run'] < 1:
                    # prevent two executors running too close to each other, setting the interval to be 2 second for now
                    self._pipeline['last_run'] = current_ts + sleep_time
                    print('sleep for 2 second')
                    time.sleep(sleep_time)
                # print('data_queue size: ' + str(self._data_queue.qsize()))
                future_result = self._loop.run_in_executor(
                    pool, self._pipeline['func'], data)
                future_result.add_done_callback(self._sending_result)
                self._future_results.add(future_result)
                # print('Running jobs: ' + str(len(self._future_results)))

    def run(self):
        print('Waiting for incoming data...')
        self._loop.create_task(self._do_computing())
        self._loop.create_task(self.wait_closed())

    def run_ws(self):
        if self._allow_ws:
            print('Start ws server on: ' + self._host + ':' + str(self._port))
            self._server = self._loop.run_until_complete(self._start_ws_server)
        else:
            print('WS server is disabled')

    async def wait_closed(self):
        while self._stopped == False:
            await asyncio.sleep(2)
        print('Stop ws servers')
        self._server.close()
        await self._server.wait_closed()
        self._ws_server_stopped = True

    async def result_stream(self):
        while True:
            value = await self._result_queue.get()
            print(value)
            yield value
            if value == 'STOP':
                print('stop in result stream')
                break

    async def _ws_handler(self, client, path):
        print('ws client is added')
        self._clients.add(client)
        try:
            async for value in self.result_stream():
                if value == 'STOP':
                    print('stop in ws handler')
                    self._stopped = True
                    break
                for client in self._clients:
                    await client.send(value)
        except websockets.exceptions.ConnectionClosed:
            print('ws client disconnected')
            self._clients.remove(client)
        finally:
            print('remaining clients: ' + str(len(self._clients)))


class ProcessorStreamManager(object):
    def __init__(self,
                 loop,
                 init_input_port=8000,
                 init_output_port=9000,
                 window_size=12.8,
                 name='processor',
                 update_rate=2,
                 number_of_windows=None):
        self._loop = loop
        self._init_input_port = init_input_port
        self._init_output_port = init_output_port
        self._input_streams = []
        self._output_streams = []
        self._window_size = window_size
        self._update_rate = update_rate
        self._merger_manager = {}
        self._name = name
        self._status = 'stopped'
        self._ready_to_stop_loop = False
        self._number_of_windows = number_of_windows
        self._window_counter = 0

    def get_ws_urls(self):
        urls = []
        for stream in self._output_streams:
            urls.append(stream.get_ws_url())
        return urls

    def change_window_size(self, window_size):
        if window_size is not None:
            self._window_size = window_size
            for stream in self._input_streams:
                stream.change_window_size(self._window_size)

    def change_update_rate(self, update_rate):
        if update_rate is not None:
            self._update_rate = update_rate
            for stream in self._input_streams:
                stream.change_update_rate(self._update_rate)

    def add_input_stream(self, host='*', port=None):
        if port is None:
            port = len(self._input_streams) + self._init_input_port
        stream = InputStream(
            self._loop,
            host=host,
            port=port,
            window_size=self._window_size,
            update_rate=self._update_rate)
        self._input_streams.append(stream)

    def add_processor_stream(self,
                             pipeline_func,
                             host='*',
                             port=None,
                             ws_server=True,
                             **kwargs):
        if port is None:
            port = len(self._output_streams) + self._init_output_port
        output_stream = ProcessorStream(
            loop=self._loop, allow_ws=ws_server, host=host, port=port)
        output_stream.set_processor_pipeline(pipeline_func, **kwargs)
        self._output_streams.append(output_stream)
        return self

    def _send_to_processor(self, chunk):
        for output_stream in self._output_streams:
            output_stream.put_input_queue(chunk)

    async def _input_chunk_handler(self, input_stream):
        async for chunk in input_stream.get_chunk_stream():
            if chunk == 'STOP':
                self._send_to_processor(chunk)
                break
            data_type = chunk.get_data_type()
            if data_type in self._merger_manager:
                merger = self._merger_manager[data_type]
            else:
                merger = ChunkMerger(
                    data_type,
                    n_streams=len(self._input_streams),
                    update_rate=self._update_rate)
                self._merger_manager[data_type] = merger
            merger.process_chunk(chunk, self._send_to_processor)

    def start_input_streams(self):
        session_st = time.time()
        print(session_st)
        for input_stream in self._input_streams:
            print('Starting input stream: ' + input_stream.get_ws_url())
            input_stream.set_session_st(session_st)
            input_stream.start()
            self._loop.create_task(self._input_chunk_handler(input_stream))

    async def auto_close(self):
        stop_processor = False
        while True:
            for stream in self._output_streams:
                if stream._result_counter >= self._number_of_windows:
                    stop_processor = True
                    break
            if stop_processor == True:
                break
            await asyncio.sleep(1)
        print('Stop in auto close')
        self.stop()

    async def wait_closed(self):
        stop = False
        while stop == False:
            stop = True
            for stream in self._output_streams:
                stop &= stream._ws_server_stopped
            await asyncio.sleep(1)
        print('Ready to stop loop')
        self._ready_to_stop_loop = True

    def stop(self):
        success = True
        for stream in self._input_streams:
            stream.stop()
        return success

    def start(self):
        for stream in self._output_streams:
            stream.run()
            stream.run_ws()
        if self._number_of_windows is not None and self._number_of_windows > 0:
            print("Start auto close function")
            self._loop.create_task(self.auto_close())
        print('Start closing waiter')
        self._loop.create_task(self.wait_closed())
        self.start_input_streams()
        self._status = 'running'

    def start_simulation(self):
        def _simulate():
            while True:
                data_buffer = []
                for i in range(50 * 5):
                    data = {
                        'HEADER_TIME_STAMP':
                        datetime.fromtimestamp(
                            time.time()).strftime('%Y-%m-%d %H:%M:%S'),
                        'X':
                        random.random(),
                        'Y':
                        random.random(),
                        'Z':
                        random.random()
                    }
                    data_buffer.append(data)
                for stream in self._output_streams:
                    stream.put_input_queue(data_buffer)
                time.sleep(5)

        Thread(target=_simulate).start()



import time
import asyncio
from pymetawear.discover import discover_devices, select_device
from pymetawear.client import MetaWearClient
import websockets
import json
import functools
from threading import Thread
import signal
import os
from .stream_package import SensorStreamPackage
from mbientlab.metawear import libmetawear
from mbientlab.metawear import cbindings
import logging


class TimestampCorrector(object):
    def __init__(self, sr, method='original'):
        self._method = method
        self._last_epoch = 0
        self._current_ts = 0
        self._sr = sr
        self._interval = 1.0 / self._sr
        self._correction = 0
        self._last_real_ts = 0

    def compute_correction(self, data, real_time):
        self._correction = real_time - data['epoch'] / 1000.0

    def _next_original(self, data):
        return data['epoch'] / 1000.0 + self._correction

    def _next_noloss(self, data, last_ts):
        if last_ts == 0:
            current_ts = self._next_original(data)
        else:
            current_ts = last_ts + self._interval
        return current_ts

    def _next_withloss(self, data, last_ts, last_epoch, real_time):
        if self._last_real_ts == 0:
            self._last_real_ts = real_time
        if abs(last_epoch - data['epoch'] / 1000.0) < self._interval:
            current_ts = self._next_noloss(data, last_ts)
            real_time = self._last_real_ts + self._interval
        else:
            next_original_ts = self._next_original(data)
            next_noloss_ts = self._next_noloss(data, last_ts)
            if next_original_ts - next_noloss_ts > 2 * self._interval:
                # late for more than two intervals, renew timestamp
                current_ts = next_original_ts
            else:
                current_ts = next_noloss_ts
        self._last_epoch = self._next_original(data)
        if abs(current_ts - real_time) >= self._interval:
            current_ts = real_time
        self._last_real_ts = real_time
        return current_ts

    def next(self, data, real_time):
        if self._method == 'original':
            self._current_ts = self._next_original(data)
        elif self._method == 'noloss':
            self._current_ts = self._next_noloss(data, self._current_ts)
        elif self._method == 'withloss':
            self._current_ts = self._next_withloss(data, self._current_ts,
                                                   self._last_epoch, real_time)
        return self._current_ts


class MetaWearStream(Thread):
    def __init__(self,
                 loop,
                 address,
                 name,
                 order,
                 host='localhost',
                 port=8000,
                 accel_sr=50,
                 accel_grange=8):
        Thread.__init__(self)
        self._device = None
        self._model_name = 'NA'
        self._accel_sr = accel_sr
        self._accel_grange = accel_grange
        self._address = address
        self._name = name
        self._order = order
        self._clients = set()
        self._host = host
        self._port = port
        self._actual_sr = 0
        self._last_minute = 0
        if loop is not None:
            self.set_loop(loop)
        self._battery_ts_corrector = TimestampCorrector(
            sr=1, method='withloss')
        self._ts_corrector_original = TimestampCorrector(
            sr=self._accel_sr, method='original')
        self._ts_corrector_noloss = TimestampCorrector(
            sr=self._accel_sr, method='noloss')
        self._ts_corrector_withloss = TimestampCorrector(
            sr=self._accel_sr, method='withloss')
        self._last_ts = 0
        self._accel_count = 0
        self._battery_count = 0
        self._stop = False
        self._status = "Available"

    def set_loop(self, loop):
        self._loop = loop
        self._accel_queue = asyncio.Queue(loop=self._loop)
        self._start_ws_server = websockets.serve(
            self._ws_handler,
            host=self._host,
            port=self._port,
            loop=self._loop)

    def get_device_address(self):
        return self._address

    def get_stream_name(self):
        return self._name

    def get_ws_uri(self):
        return 'ws://' + self._host + ':' + str(self._port)

    def run(self):
        while not self._stop:
            try:
                m = MetaWearClient(
                    str(self._address), connect=True, debug=False)
                self._device = m
                self._model_name = self.get_model_name()
            except:
                logging.info('Retry connect to ' + self._address)
                time.sleep(1)
                continue
            break
        logging.info("New metawear connected: {0}".format(m))
        # high frequency throughput connection setup
        m.settings.set_connection_parameters(7.5, 7.5, 0, 6000)
        # Up to 4dB for Class 2 BLE devices
        # https://github.com/hbldh/pymetawear/blob/master/pymetawear/modules/settings.py
        # https://mbientlab.com/documents/metawear/cpp/0/settings_8h.html#a335f712d5fc0587eff9671b8b105d3ed
        # Hossain AKMM, Soh WS. A comprehensive study of Bluetooth signal parameters for localization. 2007 Ieee 18th International Symposium on Personal, Indoor and Mobile Radio Communications, Vols 1-9. 2007:428-32.
        m.settings.set_tx_power(power=4)

        m.accelerometer.set_settings(
            data_rate=self._accel_sr, data_range=self._accel_grange)
        m.accelerometer.high_frequency_stream = True
        m.accelerometer.notifications(callback=self._accel_handler)
        m.settings.notifications(callback=self._battery_handler)
        self._status = "Stream"
        while not self._stop:
            time.sleep(1)
            m.settings.read_battery_state()
        return self

    def get_model_name(self):
        model_code = libmetawear.mbl_mw_metawearboard_get_model(
            self._device.mw.board)
        metawear_models = cbindings.Model()
        model_names = list(
            filter(lambda attr: '__' not in attr, dir(metawear_models)))
        for name in model_names:
            if getattr(metawear_models, name) == model_code:
                return name
        return 'NA'

    def get_status(self):
        return self._status

    def run_ws(self, keep_history=True):
        self._keep_history = keep_history
        logging.info('Start ws server on: ' + self.get_ws_uri())
        self._server = self._loop.run_until_complete(self._start_ws_server)

    def stop(self):
        logging.info('Disconnect from sensor')
        success = False
        try:
            self._device.accelerometer.stop()
            self._device.disconnect()
        except Exception as e:
            logging.error(str(e))
        logging.info('Stop ws server')
        try:
            self._stop = True
            self._server.close()
            self._loop.run_until_complete(self._server.wait_closed())
            self._status = "Available"
            success = True
        except Exception as e:
            logging.error(str(e))
        return success

    async def accel_stream(self):
        while True:
            value = await self._accel_queue.get()
            yield value

    async def _ws_handler(self, client, path):
        logging.info('ws client is added')
        self._clients.add(client)
        async for value in self.accel_stream():
            try:
                for client in self._clients:
                    await client.send(value)
                if self._stop:
                    break
            except websockets.exceptions.ConnectionClosed:
                logging.info('ws client disconnected')
                self._clients.remove(client)
                logging.info('remaining clients: ' + str(len(self._clients)))

    def _calibrate_accel_coord_system(self, x, y, z):
        # axis values are calibrated according to the coordinate system of Actigraph GT9X
        # http://www.correctline.pl/wp-content/uploads/2015/01/ActiGraph_Device_Axes_Link.png
        if self._model_name == 'METAMOTION_R':
            # as normal wear in the case on wrist
            calibrated_x = y
            calibrated_y = -x
            calibrated_z = z
        else:
            calibrated_x = x
            calibrated_y = y
            calibrated_z = z
        return (calibrated_x, calibrated_y, calibrated_z)

    def _pack_accel_data(self, data):
        real_ts = time.time()
        if self._accel_count == 0:
            self._ts_corrector_original.compute_correction(data, real_ts)
            self._ts_corrector_noloss.compute_correction(data, real_ts)
            self._ts_corrector_withloss.compute_correction(data, real_ts)
        package = SensorStreamPackage()
        package.set_index(self._accel_count)
        package.set_device_id(self._address)
        package.set_stream_name(self._name)
        package.set_stream_order(self._order)
        package.set_data_type('accel')
        package.add_custom_field('HEADER_TIME_STAMP_REAL', real_ts)
        package.add_custom_field(
            'HEADER_TIME_STAMP_ORIGINAL',
            self._ts_corrector_original.next(data, real_ts))
        package.add_custom_field('HEADER_TIME_STAMP_NOLOSS',
                                 self._ts_corrector_noloss.next(data, real_ts))
        package.set_timestamp(self._ts_corrector_withloss.next(data, real_ts))
        value = {}
        calibrated_values = self._calibrate_accel_coord_system(
            data['value'].x, data['value'].y, data['value'].z)

        value['X'] = calibrated_values[0]
        value['Y'] = calibrated_values[1]
        value['Z'] = calibrated_values[2]
        package.set_value(value)

        return package.to_json_string(), package

    def _pack_battery_data(self, data):
        real_ts = time.time()
        if self._battery_count == 0:
            self._battery_ts_corrector.compute_correction(data, time.time())
        package = SensorStreamPackage()
        package.set_index(self._battery_count)
        package.set_device_id(self._address)
        package.set_stream_name(self._name)
        package.set_stream_order(self._order)
        package.set_data_type('battery')
        package.set_timestamp(self._battery_ts_corrector.next(data, real_ts))
        value = {}
        value['BATTERY_VOLTAGE'] = data['value'].voltage
        value['BATTERY_PERCENTAGE'] = data['value'].charge
        package.set_value(value)
        return package.to_json_string()

    def _accel_handler(self, data):
        self._status = "Stream"
        json_package, package = self._pack_accel_data(data)
        if package._package['HEADER_TIME_STAMP_ORIGINAL'] < self._last_ts:
            logging.warning('earlier sample detected')
        else:
            self._last_ts = package._package['HEADER_TIME_STAMP_ORIGINAL']

        if self._keep_history:
            self._loop.call_soon_threadsafe(self._accel_queue.put_nowait,
                                            json_package)
        else:
            if len(self._clients) > 0:
                self._loop.call_soon_threadsafe(self._accel_queue.put_nowait,
                                                json_package)
        if self._last_minute == 0:
            logging.debug(package.to_dataframe())
            self._last_minute = time.time() * 1000.0
        if time.time() * 1000.0 - self._last_minute > 1000.0:
            logging.info(self._address + ' sr: ' + str(self._actual_sr) +
                         ', ' + 'time diff: ' +
                         str(package.get_timestamp() -
                             package._package['HEADER_TIME_STAMP_REAL']))
            self._actual_sr = 0
            self._last_minute = time.time() * 1000.0
            self._loop.call_soon_threadsafe(
                logging.debug, self._address + ' queue size: ' + str(
                    self._accel_queue.qsize()))
        else:
            self._actual_sr = self._actual_sr + 1
        self._accel_count += 1

    def _battery_handler(self, data):
        battery = data['value']
        self._loop.call_soon_threadsafe(
            logging.info, 'battery level: ' + str(battery.voltage) + ', ' +
            str(battery.charge))
        json_package = self._pack_battery_data(data)
        if self._keep_history:
            self._loop.call_soon_threadsafe(self._accel_queue.put_nowait,
                                            json_package)
        else:
            if len(self._clients) > 0:
                self._loop.call_soon_threadsafe(self._accel_queue.put_nowait,
                                                json_package)
        self._battery_count += 1


class MetaWearScanner():
    def scan(self):
        metawears = set()
        retries = 0
        while retries < 3:
            logging.info('Scanning metawear devices nearby...')
            try:
                retries += 1
                candidates = discover_devices(timeout=5)
                metawears |= set(
                    map(lambda d: d[0],
                        filter(lambda d: d[1] == 'MetaWear', candidates)))
            except ValueError as e:
                logging.error(str(e))
                continue
        return list(metawears)


class MetaWearStreamManager(object):
    def __init__(self, max_devices=2, init_port=8000):
        self._max_devices = max_devices
        self._init_port = init_port
        self.streams = []
        self._loop = asyncio.new_event_loop()

    def reset_ble_adaptor(self):
        os.system('radiocontrol.exe Bluetooth off')
        time.sleep(1)
        os.system('radiocontrol.exe Bluetooth on')

    def _scan_for_metawears(self):
        metawears = []
        metawear_stream_names = []
        while len(metawears) < self._max_devices:
            logging.info('Scanning metawear devices nearby...')
            try:
                addr = select_device(timeout=3)
                stream_name = input('enter stream name: ')
                metawears.append(addr)
                metawear_stream_names.append(stream_name)
            except ValueError as e:
                continue
            # metawears = list(filter(lambda d: d[1] == 'MetaWear', devices))
            time.sleep(1)
        return metawears, metawear_stream_names

    def _autoscan_for_metawears(self):
        metawears = set()
        retries = 0
        while len(metawears) < self._max_devices and retries <= 5:
            logging.info('Scanning metawear devices nearby...')
            try:
                retries += 1
                candidates = discover_devices(timeout=3)
                metawears |= set(
                    map(lambda d: d[0],
                        filter(lambda d: d[1] == 'MetaWear', candidates)))
            except ValueError as e:
                continue
        return metawears

    def get_status(self, addr):
        for stream in self.streams:
            if stream.get_device_address() == addr:
                return stream.get_status()
        return "Unknown"

    def get_stream_name(self, addr):
        for stream in self.streams:
            if stream.get_device_address() == addr:
                return stream.get_stream_name()
        return None

    def get_sr(self, addr):
        for stream in self.streams:
            if stream.get_device_address() == addr:
                return stream._accel_sr
        return None

    def get_grange(self, addr):
        for stream in self.streams:
            if stream.get_device_address() == addr:
                return stream._accel_grange
        return None

    def get_order(self, addr):
        for stream in self.streams:
            if stream.get_device_address() == addr:
                return stream._order
        return None

    def get_ws_url(self, addr):
        for stream in self.streams:
            if stream.get_device_address() == addr:
                return stream.get_ws_uri()
        return None

    def disconnect(self, addr):
        to_remove = None
        success = False
        for stream in self.streams:
            if stream.get_device_address() == addr:
                success = stream.stop()
                to_remove = stream
                break
        if to_remove is not None:
            if success:
                self.streams.remove(to_remove)
                return "Success"
            else:
                return "Failure"
        else:
            logging.info("Can not find the stream with addr: " + addr)
            return "Not found"

    def scan(self):
        metawears = self._autoscan_for_metawears()
        logging.debug(metawears)
        return metawears

    def stream_exists(self, stream):
        for s in self.streams:
            if stream.get_device_address() == s.get_device_address():
                return 'Address in use'
            elif stream.get_stream_name() == s.get_stream_name():
                return 'Name in use'
            elif stream._order == s._order:
                return 'Order in use'
        return False

    def init(self, metawears, loop):
        self._loop = loop
        results = metawears
        for metawear in metawears:
            i = metawears.index(metawear)
            addr = metawear['address']
            name = metawear['name']
            order = metawear['order']
            accel_sr = metawear['sr']
            accel_grange = metawear['grange']
            port = self._init_port + order
            stream = MetaWearStream(
                self._loop,
                addr,
                name,
                order,
                port=port,
                accel_sr=accel_sr,
                accel_grange=accel_grange)
            error_code = self.stream_exists(stream)
            logging.debug(error_code)
            if error_code == False:
                logging.info('Start to run a sensor')
                self.streams.append(stream)
                stream.run_ws(keep_history=True)
                try:
                    stream.start()
                    results[i]['ws_uri'] = stream.get_ws_uri()
                except KeyboardInterrupt:
                    stream.stop()
                    results[i]['ws_uri'] = 'failed to start'
                # rest for a second before connecting to the next stream
                time.sleep(1)
            else:
                logging.warning('error in running sensor')
                results[i]['error_code'] = error_code
        return results

    def start(self,
              accel_sr=50,
              accel_grange=8,
              ws_server=True,
              keep_history=True):
        self.reset_ble_adaptor()
        metawears, metawear_stream_names = self._scan_for_metawears()
        logging.debug(metawears)
        logging.debug(metawear_stream_names)

        for i in range(self._max_devices):
            address = metawears[i]
            # address = 'D5:D2:01:2B:E7:6D'
            name = metawear_stream_names[i]
            port = self._init_port + i
            stream = MetaWearStream(
                self._loop,
                address,
                name,
                i,
                port=port,
                accel_sr=accel_sr,
                accel_grange=accel_grange)
            self.streams.append(stream)
            # start ws server first
            if ws_server:
                stream.run_ws(keep_history=keep_history)
            try:
                stream.start()
            except KeyboardInterrupt:
                stream.stop()
            # rest for a second before connecting to the next stream
            time.sleep(1)

        if ws_server:
            signal.signal(signal.SIGINT, signal.SIG_DFL)
            self._loop.run_forever()



from enum import Enum
import json
import pandas as pd


class SensorStreamChunk(object):
    def __init__(self, stream_order, stream_name, device_id, data_type):
        self._chunk = {
            'LAST_CHUNK': None,
            'START_TIME': None,
            'STOP_TIME': None,
            'STREAM_NAME': stream_name,
            'STREAM_ORDER': stream_order,
            'DEVICE_ID': device_id,
            'PACKAGES': [],
            'DATA_TYPE': data_type,
            'CHUNK_INDEX': 0,
            'LAST_CHUNK_ST': None
        }

    def get_chunk_size(self):
        return len(self._chunk['PACKAGES'])

    def add_package(self, package):
        self._chunk['PACKAGES'].append(package)

    def set_packages(self, packages):
        self._chunk['PACKAGES'] = packages

    def get_packages(self):
        return self._chunk['PACKAGES']

    def get_data_type(self):
        return self._chunk['DATA_TYPE']

    def get_stream_name(self):
        return self._chunk['STREAM_NAME']

    def get_stream_order(self):
        return self._chunk['STREAM_ORDER']

    def get_device_id(self):
        return self._chunk['DEVICE_ID']

    def get_chunk_index(self):
        return self._chunk['CHUNK_INDEX']

    def increment_chunk_index(self):
        self._chunk['CHUNK_INDEX'] += 1

    def set_data_type(self, data_type):
        self._chunk['DATA_TYPE'] = data_type

    def set_chunk_st(self, chunk_st):
        self._chunk['START_TIME'] = chunk_st

    def set_chunk_et(self, chunk_et):
        self._chunk['STOP_TIME'] = chunk_et

    def get_chunk_st(self):
        return self._chunk['START_TIME']

    def get_chunk_et(self):
        return self._chunk['STOP_TIME']

    def set_last_chunk_st(self, last_chunk_st):
        self._chunk['LAST_CHUNK_ST'] = last_chunk_st

    def get_last_chunk_st(self):
        return self._chunk['LAST_CHUNK_ST']

    def to_json_string(self):
        return json.dumps(self._chunk)

    def __str__(self):
        return self.get_stream_name() + ' ' + self.get_data_type(
        ) + " chunk " + str(self.get_chunk_index()) + " size: " + str(
            len(self.get_packages())) + ", start time: " + str(
                self.get_chunk_st()) + ', stop time: ' + str(
                    self.get_chunk_et())

    def from_json_string(self, json_chunk):
        self._chunk = json.loads(json_chunk)


class SensorStreamPackage(object):
    def __init__(self):
        self._package = {
            'HEADER_TIME_STAMP': None,
            'DATA_TYPE': None,
            'DEVICE_ID': None,
            'STREAM_NAME': None,
            'VALUE': {},
            'PACKAGE_INDEX': 0,
            'STREAM_ORDER': None
        }

    def set_index(self, index):
        self._package['PACKAGE_INDEX'] = index

    def get_value_entries(self):
        return self._package['VALUE'].keys()

    def get_data_type(self):
        return self._package['DATA_TYPE']

    def get_timestamp(self):
        return self._package['HEADER_TIME_STAMP']

    def get_device_id(self):
        return self._package['DEVICE_ID']

    def set_timestamp(self, timestamp):
        self._package['HEADER_TIME_STAMP'] = timestamp

    def set_value(self, value):
        self._package['VALUE'] = value

    def set_device_id(self, device_id):
        self._package['DEVICE_ID'] = device_id

    def set_stream_name(self, stream_name):
        self._package['STREAM_NAME'] = stream_name

    def set_stream_order(self, stream_order):
        self._package['STREAM_ORDER'] = stream_order

    def get_stream_name(self):
        return self._package['STREAM_NAME']

    def get_stream_order(self):
        return self._package['STREAM_ORDER']

    def set_data_type(self, data_type):
        self._package['DATA_TYPE'] = data_type

    def add_custom_field(self, name, value):
        self._package[name] = value

    def to_json_string(self):
        return json.dumps(self._package)

    def from_json_string(self, json_package):
        self._package = json.loads(json_package)

    def to_dataframe(self):
        df_dict = {}
        value_dict = None
        for name, value in self._package.items():
            if name == 'VALUE':
                value_dict = value
            else:
                df_dict[name] = value
        df_dict = {**df_dict, **value_dict}
        return pd.DataFrame.from_records([df_dict])



from setuptools import setup, find_packages

setup(
    name='padar_realtime',
    version='1.4.2',
    packages=find_packages(),
    include_package_data=True,
    description=
    'Real-time engine used to process and visualize accelerometer data for padar package',
    long_description=open('README.md').read(),
    install_requires=[
        "websockets", "flask", "arrow", "pymetawear", "pandas", 'metawear',
        'warble'
    ],
)


import asyncio
import websockets
from threading import Thread
from .stream_package import SensorStreamPackage, SensorStreamChunk
import math
import copy


class InputStream(Thread):
    def __init__(self,
                 loop,
                 host,
                 port,
                 window_size,
                 update_rate,
                 session_st=None):
        Thread.__init__(self)
        self._host = host
        self._port = port
        self._loop = loop
        self._queue = asyncio.Queue(loop=self._loop)
        self._window_size = window_size
        self._update_rate = update_rate
        self._chunk_manager = {}
        self._session_st = session_st
        self._stop = False

    def get_ws_url(self):
        return 'ws://' + self._host + ":" + str(self._port)

    def change_window_size(self, window_size):
        self._window_size = window_size

    def change_update_rate(self, update_rate):
        self._update_rate = update_rate

    def set_session_st(self, session_st):
        self._session_st = session_st

    def _create_input_package(self, value):
        input_package = SensorStreamPackage()
        input_package.from_json_string(value)
        return input_package

    def stop(self):
        self._stop = True

    def send_stop_to_queue(self):
        self._loop.call_soon_threadsafe(self._queue.put_nowait, 'STOP')

    async def _ws_handler(self):
        while True:
            try:
                async with websockets.connect(self.get_ws_url()) as websocket:
                    try:
                        print('connected to ' + self.get_ws_url())
                        while True:
                            data = await websocket.recv()
                            if self._stop:
                                self.send_stop_to_queue()
                                break
                            self._handle_input_stream(data)
                    except websockets.exceptions.ConnectionClosed as e:
                        print(str(e))
                        if self._stop:
                            self.send_stop_to_queue()
                            break
                    except Exception as e:
                        print(str(e))
                        if self._stop:
                            self.send_stop_to_queue()
                            break
            except OSError as e:
                print(str(e))
                asyncio.sleep(3)
                print('Retry connect to ' + 'ws://' + self._host + ':' +
                      str(self._port))
                if self._stop:
                    self.send_stop_to_queue()
                    break
            except Exception as e:
                self.send_stop_to_queue()
                break

    async def get_chunk_stream(self):
        while True:
            chunk = await self._queue.get()
            yield chunk
            if chunk == 'STOP':
                break

    def _handle_input_stream(self, data):
        package = self._create_input_package(data)
        data_type = package.get_data_type()
        device_id = package.get_device_id()
        stream_name = package.get_stream_name()
        stream_order = package.get_stream_order()
        if data_type not in self._chunk_manager:
            chunk = SensorStreamChunk(stream_order, stream_name, device_id,
                                      data_type)
        else:
            chunk = self._chunk_manager[data_type]

        # set start time of current chunk
        self.__set_chunk_st(chunk)

        # set stop time of current chunk
        self.__set_chunk_et(chunk)

        # process package
        self.__process_package(chunk, package)

        self._chunk_manager[data_type] = chunk

    def __set_chunk_st(self, chunk):
        if chunk.get_chunk_st() is None and chunk.get_last_chunk_st() is None:
            chunk.set_chunk_st(self._session_st)
        elif chunk.get_chunk_st() is None and chunk.get_last_chunk_st(
        ) is not None:
            chunk.set_chunk_st(chunk.get_last_chunk_st() + self._update_rate)

    def __set_chunk_et(self, chunk):
        if chunk.get_chunk_et() is None:
            chunk.set_chunk_et(chunk.get_chunk_st() + self._window_size)

    def __process_package(self, chunk, input_package):
        if input_package.get_timestamp() < chunk.get_chunk_st():
            # when the input data has a timestamp from the past, out-of-date data
            # print('This package is from past, do not add to current chunk')
            return
        elif input_package.get_timestamp() >= chunk.get_chunk_st(
        ) and input_package.get_timestamp() < chunk.get_chunk_et():
            # when the input data belongs to current chunk
            chunk.add_package(input_package)
        else:  # input_package.get_timestamp() >= chunk.get_chunk_et()
            # when the input data belongs to the next chunk
            copied_chunk = copy.deepcopy(chunk)

            # keep only partial of current chunk for the next turn
            keep_packages = list(
                filter(
                    lambda p: p.get_timestamp() > chunk.get_chunk_st() + self._update_rate,
                    chunk.get_packages()))

            # add current package to the chunk
            chunk.set_packages(keep_packages)
            chunk.add_package(input_package)

            # set new chunk start time and clear old ones
            chunk.set_last_chunk_st(chunk.get_chunk_st())
            chunk.set_chunk_st(None)
            chunk.set_chunk_et(None)
            chunk.increment_chunk_index()

            # send chunk to queue
            self._loop.call_soon_threadsafe(self._queue.put_nowait,
                                            copied_chunk)

    def run(self):
        self._loop.create_task(self._ws_handler())
